local MAJOR, MINOR = "Xii_LibEnv-1.0", 2
local lib = LibStub:NewLibrary(MAJOR, MINOR)


local assert, ipairs, pairs, setmetatable, table, type, unpack, rawget, rawset = assert, ipairs, pairs, setmetatable, table, type, unpack, rawget, rawset

if not lib then
    return    -- already loaded and no upgrade necessary
end

local mtWeakKeyValue = { __mode = "kv" }

local Cache = setmetatable({}, { __mode = "k" })
local Namespaces = {}

local mt = {
    __index = function(t, k)
        if (Cache[t][k]) then
            -- print("Cached!", t, k)
            return Cache[t][k]
        end

        local v = _G[k]
        if (v) then
            -- print("Miss!", t, k)
            Cache[t][k] = v
            return v
        end
    end
}

lib.Cache = Cache
lib.Namespaces = Namespaces

local mixin
mixin = function(t, ...)
    print("mixin")
    local argv = {...}

    if (#argv > 1) then
        for i,v in ipairs(argv) do
            mixin(t, v)
        end
    else
        for k,v in pairs(argv[1]) do
            if (type(k) == "number" and type(v) == "table") then
                mixin(t, v)
            else
                t[k] = v
            end
        end
    end

    return t
end


function lib:Use(name, ...)
    local env = self:Get(name, ...)
    setfenv(2, env)

    return env
end


function lib:Mixin(envOrName, ...)
    local env

    if (type(envOrName) == "string") then
        env = self:Get(envOrName)
    elseif (type(envOrName) == "table") then
        env = envOrName
    end

    return mixin(env, ...)
end


function lib:Get(name)
    local env = Namespaces[name]

    if (not env) then
        env = setmetatable({
            _G = _G,
            rawget = rawget,
            rawset = rawset,
            print = print,
        }, mt)

        Cache[env] = setmetatable({}, mtWeakKeyValue)
        Namespaces[name] = env
    end

    return env
end
